#!/bin/bash
# backup.sh
# 
# Backup up Unix-like computers
# 
# itk@samfundet.no
#
# This script was first checked into RCS in 2000, and has since backed up a few
# servers almost every day, whenever not broken.
# 
# It's a bit peculiar, but quite effective. The following is worth keeping in
# mind when hacking:
#
# Because of the way we run this script from cron, informational logging goes
# to stderr and warnings and errors to stdout. This way, we get mail whenever
# stuff fails, and other output is redirected to the log file.
# 
# The script is halfway Norwegian and halfway English. Newer modifications are
# in English, we should probably stick to that.

# The computer to backup.
computer=$1

# Configuration
LOCKFILE=/home/backup/backuprun.lockfile.$computer
confdir=/home/backup/conf/              # Configuration files
storagedir=/backup                      # Where we keep backups
maxnumfull=3                            # Number of full backups
daysbetweenfull=30                      # Days between full backups
daysbetweenfullforfixed=32              # Days between full backups for machines with fixed
                                        # full-backup date (in case the fixed day gets missed)
DAY_OF_MONTH=`date "+%d" | sed s/^0//`  # from 1 to 31
DATE=`date "+%Y%m%d%H%M"`               #format: touch
DATEs=`date "+%Y-%m-%d %H:%M"`          #format: tar

# Exclude-pattern
exclude=$confdir/exclude
[ ! -f $exclude ] && printf "tmp\ncore\n" > $exclude

if [ -z "$computer" ]; then
	echo "Usage: $0 COMPUTER"
	exit 1
fi

# Die more or less gracefully (at least notify user)
die() {
	echo `date`": Something nasty happened."
	rm $LOCKFILE
	exit 1
}
diemsg() {
	echo `date`": $computer: $@"
	die
}

# Trap C-c and kill
trap die SIGINT SIGTERM

# Don't start if we're already running
if [ -e $LOCKFILE ]; then
	echo `date`": $LOCKFILE exists for pid `cat $LOCKFILE`, exiting."
	echo `date`": $LOCKFILE exists for pid `cat $LOCKFILE`, exiting." >&2
	exit 1
fi;
touch $LOCKFILE
echo $$ > $LOCKFILE

echo `date`": Backup run starting" >&2

umask 027


backup()
{
	echo -n `date` >&2
	printf " $computer: $computer:$filesystem $backuplevel backup\n" >&2

	SNARFILE="${storagedir}/${computer}/${sfilesystem}/.incremental.snar"
	incrementalsnar="/root/.backup/${sfilesystem}.snar"
	if [ "$backuplevel" = "daglig" ]
	then
		# If incremental backup, we need to copy the incremental status to $computer.
		# If it does not exist, whine a bit and then run date-based instead.
		if [ -s "$SNARFILE" ]; then
			if ! scp $SNARFILE root@$computer:$incrementalsnar; then
				diemsg "Could not copy .incremental.snar to $computer"
			fi
			lastcmd="--listed-incremental=$incrementalsnar --no-check-device"
		else
			echo `date`" $computer: Missing incremental.snar for $filesystem, doing date-based backup instead until next full backup" >&2
			incrementalsnar=""
			lastd=`cat ../.date`
			lastcmd="--newer='$lastd'"
		fi
	else
		lastcmd="--listed-incremental=$incrementalsnar"
	fi

	# We try to run tar on the remote computer
	#    c create archive
	#    C change to directory first
	#    . where to start taring (see C)
	#    $lastcmd only files newer than this
	#    --one-file-system don't traverse file systems
	#    --use-compress-program compress it using gzip or pigz
	#    --exclude-from file to get exclusion pattern from
	#    Pipe the stuff over ssh to ourselves, run pee to cat the contents to a
	#    file and build a file list, respectivly.  
	TARFILE=$DATE.tmp
	TARCMD="ssh root@$computer \"nice -n 19 ionice -c3 tar --one-file-system --use-compress-program $compressor -cf - -C $filesystem . $lastcmd \
		--exclude-from=/root/.backup/exclude\" | pee \"cat > $TARFILE\" \"tar tzvf -\""
	echo `date`" $computer: Running $TARCMD" >&2
	eval $TARCMD > $DATE.idx

	if [ $? -eq 0 ] && [ -s $TARFILE ]; then
		# File is >0 in size and neither cat or tar tzvf failed; we assume it worked.
 
		if [ "$incrementalsnar" ]; then
			if ! scp root@$computer:$incrementalsnar $SNARFILE.tmp; then
				diemsg "Could not copy .incremental.snar from $computer"
			fi
		fi

		# Move tar file in place
		mv $TARFILE $DATE.tgz
		mv $SNARFILE.tmp $SNARFILE

		# Update timestamp
		echo $DATEs > ../.date && touch -t $DATE ../.date

		# Make a sorted file list as well
		sort -k6 < $DATE.idx > $DATE.sdx

		# Fix permissions
		chmod 600 *tgz                      #only for us
		chmod 644 *sdx *idx 2>/dev/null     #everyone can read

		# Let the remote computer know that we ran a successful backup (for nagios)
		ssh root@$computer touch $filesystem/.lastbackup
	else
		# Something wrong happened.
		rm $TARFILE
		diemsg "tar failed or $TARFILE empty. $backuplevel backup of $computer:$filesystem failed and deleted"
	fi
}

# Check that the target filesystem is mounted (actually check that it's not
# the root filesystem)
if [ ! -d "$storagedir/$computer" ]; then
	diemsg "Target filesystem ($storagedir/$computer) does not exist. Aborting"
fi
rootfilesystem=`df -P /`
targetfilesystem=`df -P "$storagedir/$computer"`
if [ "$rootfilesystem" == "$targetfilesystem" ]; then
	diemsg "Target filesystem ($storagedir/$computer) was mounted on /. Aborting"
fi

echo `date`" $computer: Backing up $computer" >&2

# Try to SSH to the computer without entering a password.
if ! `ssh -n -o NumberOfPasswordPrompts=0 root@$computer /bin/true`; then
	diemsg "Could not use passwordless SSH."
fi

# Check if pigz is available
if ssh -n root@$computer "pigz -V 2>/dev/null"; then
	compressor=pigz
else
	echo `date`" $computer: pigz missing; falling back to gzip."
	compressor=gzip
fi

# Check dump bit in fstab
filesystems=`ssh -n root@$computer "cat /etc/fstab" \
	| grep -v nfs \
	| grep -v "^#" \
	| grep -v "^$" \
	| awk '{ if ( $(NF-1) != "0" ) print $2}' `

# Clean up our dir at this client
if ! ssh root@$computer "rm -r /root/.backup ; mkdir -m 700 /root/.backup"; then
	diemsg "Could not create backup staging area at $computer:/root/.backup"
fi

# Try to copy $exclude to $computer
if ! scp $exclude root@$computer:/root/.backup/exclude > /dev/null; then
	diemsg "Could not copy exclude.txt to $computer"
fi

# Try to copy preeexec and postexec if they exist
if [ -f $confdir/preexec.$computer ]; then
	if ! scp $confdir/preexec.$computer  root@$computer:/root/.backup/preexec >&2; then
		diemsg "Could not copy preexec.$computer to $computer:/root/.backup/preexec"
	fi
fi
if [ -f $confdir/postexec.$computer ]; then
	if ! scp $confdir/postexec.$computer root@$computer:/root/.backup/postexec >&2; then
		diemsg "Could not copy postexec.$computer to $computer:/root/.backup/postexec"
	fi
fi

# Try to run preexec if it exists
if ! ssh root@$computer "[ ! -f /root/.backup/preexec ] || /bin/bash -x /root/.backup/preexec" >&2; then
	diemsg "Could not run $computer:/root/.backup/preexec"
fi

for filesystem in $filesystems; do
	# Remove / and $ (for Windows) for path names
	sfilesystem=`echo $filesystem | tr '\/\$' '__'`

	# Prepare storage area
	mkdir -m 755 -p $storagedir/$computer/$sfilesystem/full 2>/dev/null
	mkdir -m 755 -p $storagedir/$computer/$sfilesystem/daglig 2>/dev/null

	# Default backuplevel
	backuplevel=daglig

	if [ ! -s $storagedir/$computer/$sfilesystem/.date ]
	then
		# Take the first full backup of this filesystem on this computer
		backuplevel=full
		echo $DATEs > $storagedir/$computer/$sfilesystem/.date
	fi

	# Check if we want a full backup
	if [ -f $confdir/fastfullbackupdag.$computer.$sfilesystem ]; then
		fullbackup_min_for_this_machine=$daysbetweenfullforfixed
		if [ "$DAY_OF_MONTH" = "`cat $confdir/fastfullbackupdag.$computer.$sfilesystem`" ]; then
			backuplevel=full
		fi
	elif [ -f $confdir/fastfullbackupdag.$computer ]; then
		fullbackup_min_for_this_machine=$daysbetweenfullforfixed
		if [ "$DAY_OF_MONTH" = "`cat $confdir/fastfullbackupdag.$computer`" ]; then
			backuplevel=full
		fi
	else
		fullbackup_min_for_this_machine=$daysbetweenfull 
	fi

	if [ -z "`find $storagedir/$computer/$sfilesystem/full/ -name \*tgz -mtime -$fullbackup_min_for_this_machine`" ]; then
		backuplevel=full
	fi

	# We want to be here
	cd $storagedir/$computer/$sfilesystem/$backuplevelmsg || die "$storagedir/$computer/$sfilesystem/$backuplevel does not exist"

	# Perform the actual backup
	backup

	# Check if this box has a custom number of full backups
	if [ -f $confdir/maksfulle.$computer.$sfilesystem ]; then
		mf=$((`cat $confdir/maksfulle.$computer`+1))
	elif [ -f $confdir/maksfulle.$computer ] ; then
		mf=$((`cat $confdir/maksfulle.$computer`+1))
	else
		mf=$(($maxnumfull+1))
	fi

	# Delete old full backups
	for full in `ls -1t $storagedir/$computer/$sfilesystem/full/*tgz | tail -n +$mf`; do
		prefix=`echo $full | sed "s/\.[^.]*$//"`
		echo `date`": $computer:$filesystem Deleting full backup $prefix" >&2
		rm $prefix*
	done

	# Delete incremental backups older than the oldest full backup
	oldf=`ls -t1 $storagedir/$computer/$sfilesystem/full/*tgz | tail -1`
	find \
		$storagedir/$computer/$sfilesystem/daglig \
		-type f \
		\! -newer $oldf \
		-printf "`date`: $computer: Deleting old incremental backup: %p\n" \
		-exec rm {} \; >&2
done

# Try to run postexec if it exist
if ! ssh root@$computer "[ ! -f /root/.backup/postexec ] || /bin/bash -x /root/.backup/postexec" >&2; then
	diemsg "Could not run $computer:/root/.backup/postexec"
fi

# Remove lockfile
rm $LOCKFILE
